@@ -135,6 +135,22 @@ class Miriad(UVData):
             "xtsys",
             "ytsys",
             "baseline",
+            "obspa",
+            "antaz",
+            "antel",
+            "axisoff",
+            "cable",
+            "dazim",
+            "delev",
+            "jyperka",
+            "phaselo1",
+            "phaselo2",
+            "phasem1",
+            "themt",
+            "wfreq",
+            "wwidth",
+            "wsystemp",
+            "wcorr",
         ]
 
         extra_miriad_variables = []
@@ -169,11 +185,11 @@ class Miriad(UVData):
             # Channel widths are described per spw, just need to expand it out to be
             # for each frequency channel.
             self.channel_width = (
-                np.array(
-                    [
-                        [np.abs(chan_width)] * nchan
+                np.concatenate(
+                    tuple(
+                        np.abs(chan_width) + np.zeros(nchan)
                         for (chan_width, nchan) in zip(uv["sdf"] * 1e9, uv["nschan"])
-                    ]
+                    )
                 )
                 .flatten()
                 .astype(np.float64)
@@ -181,28 +197,24 @@ class Miriad(UVData):
             # Now setup frequency array
             # TODO: Spw axis to be collapsed in future release
             self.freq_array = np.reshape(
-                np.array(
-                    [
+                np.concatenate(
+                    tuple(
                         chan_width * np.arange(nchan) + sfreq
                         for (chan_width, nchan, sfreq) in zip(
                             uv["sdf"] * 1e9, uv["nschan"], uv["sfreq"] * 1e9
                         )
-                    ]
+                    )
                 )
                 .flatten()
                 .astype(np.float64),
                 (1, -1),
             )
             # TODO: Fix this to capture unsorted spectra
-            self.flex_spw_id_array = (
-                np.array(
-                    [
-                        [idx] * nchan
-                        for (idx, nchan) in zip(range(self.Nspws), uv["nschan"])
-                    ]
+            self.flex_spw_id_array = np.concatenate(
+                tuple(
+                    idx + np.zeros(nchan, dtype=int)
+                    for (idx, nchan) in zip(range(self.Nspws), uv["nschan"])
                 )
-                .flatten()
-                .astype(np.int64)
             )
         else:
             self.freq_array = np.reshape(
@@ -220,7 +232,7 @@ class Miriad(UVData):
         if "visunits" in uv.vartable.keys():
             self.vis_units = uv["visunits"].replace("\x00", "")
         else:
-            self.vis_units = "UNCALIB"  # assume no calibration
+            self.vis_units = "uncalib"  # assume no calibration
         if "instrume" in uv.vartable.keys():
             self.instrument = uv["instrume"].replace("\x00", "")
         else:
@@ -265,6 +277,11 @@ class Miriad(UVData):
 
         latitude = uv["latitud"]  # in units of radians
         longitude = uv["longitu"]
+
+        # Catch a weird case where where sometimes long is wrapped like RA (0 -> 2pi
+        # instead of -pi -> pi)
+        if longitude > np.pi:
+            longitude -= 2 * np.pi
         try:
             altitude = uv["altitude"]
             self.telescope_location_lat_lon_alt = (latitude, longitude, altitude)
@@ -710,7 +727,11 @@ class Miriad(UVData):
         check_extra=True,
         run_check_acceptability=True,
         strict_uvw_antpos_check=False,
-        skip_extra_sources=False,
+        calc_lst=True,
+        fix_old_proj=False,
+        fix_use_ant_pos=True,
+        check_autos=True,
+        fix_autos=True,
     ):
         """
         Read in data from a miriad file.
@@ -777,10 +798,24 @@ class Miriad(UVData):
         strict_uvw_antpos_check : bool
             Option to raise an error rather than a warning if the check that
             uvws match antenna positions does not pass.
-        skip_extra_sources : bool
-            As multiple sources are not (yet) supported, this will simply allow
-            the reader to read just the first source in the dataset, rather than
-            throwing an exception. Default is false.
+        calc_lst : bool
+            Recalculate the LST values that are present within the file, useful in
+            cases where the "online" calculate values have precision or value errors.
+            Default is True.
+        fix_old_proj : bool
+            Applies a fix to uvw-coordinates and phasing, assuming that the old `phase`
+            method was used prior to writing the data, which had errors of the order of
+            one part in 1e4 - 1e5. See the phasing memo for more details.
+        fix_use_ant_pos : bool
+            If setting `fix_old_proj` to True, use the antenna positions to derive the
+            correct uvw-coordinates rather than using the baseline vectors. Default is
+            True.
+        check_autos : bool
+            Check whether any auto-correlations have non-zero imaginary values in
+            data_array (which should not mathematically exist). Default is True.
+        fix_autos : bool
+            If auto-correlations with imaginary values are found, fix those values so
+            that they are real-only in data_array. Default is False.
 
         Raises
         ------
@@ -809,14 +844,16 @@ class Miriad(UVData):
             check_variables,
         ) = self._read_miriad_metadata(uv, correct_lat_lon=correct_lat_lon)
 
+        # update filename attribute
+        basename = filepath.rstrip("/")
+        self.filename = [os.path.basename(basename)]
+        self._filename.form = (1,)
+
         if not read_data:
             # don't read in the data. This means the object is incomplete,
             # but that may not matter for many purposes.
             return
 
-        # read through the file and get the data
-        _source = uv["source"]  # check source of initial visibility
-
         history_update_string = "  Downselected to specific "
         n_selects = 0
 
@@ -975,7 +1012,7 @@ class Miriad(UVData):
                     pol_list.append(p)
             # check not empty
             if len(pol_list) == 0:
-                raise ValueError(f"No polarizations in data matched {polarizations}")
+                raise ValueError("No polarizations in data matched input")
             if n_selects > 0:
                 history_update_string += ", polarizations"
             else:
@@ -988,7 +1025,20 @@ class Miriad(UVData):
 
         data_accumulator = {}
         pol_list = []
-        warn_extra_sources = True
+        app_ra = None
+        app_dec = None
+        frame_pa = None
+        epoch = None
+        sou_dict = {}
+        Nphase = 0
+        record_epoch = "epoch" in uv.vartable.keys()
+        record_app = ("obsra" in uv.vartable.keys()) and (
+            "obsdec" in uv.vartable.keys()
+        )
+        record_pa = "obspa" in uv.vartable.keys()
+
+        # dra_list - Stubbing out for later
+        # ddec_list - Stubbing out for later
         for (uvw, t, (i, j)), d, f in uv.all_data(raw=True):
             # control for the case of only a single spw not showing up in
             # the dimension
@@ -1002,7 +1052,7 @@ class Miriad(UVData):
             # this array should ever _not_ be 1D, but the below does cast the shape of
             # the data array correctly so that the vestigial spw-axis is preserved. At
             # some point, the below will need to be fixed -- I'm keeping this here so
-            # that I can skip reading through the MIRIAD programmers guide yet aagain.
+            # that I can skip reading through the MIRIAD programmers guide yet again.
             if len(d.shape) == 1:
                 d.shape = (1,) + d.shape
 
@@ -1014,48 +1064,71 @@ class Miriad(UVData):
                 cnt = np.ones(d.shape, dtype=np.float64)
             ra = uv["ra"]
             dec = uv["dec"]
-            # NOTE: Using our lst calculator, which uses astropy,
-            # instead of _miriad values which come from pyephem.
-            # The differences are of order 5 seconds.
-            # To use the values from the file you'd want: lst = uv['lst']
+            if record_epoch:
+                epoch = uv["epoch"]
+            if record_app:
+                app_ra = uv["obsra"]
+                app_dec = uv["obsdec"]
+            if record_pa:
+                frame_pa = uv["obspa"]
+            # LST is pulled from the file here, atlhough as some PAPER/early HERA
+            # data calculated LST from pyephem (which is inconsistent w/ astropy
+            # to the order of ~5 seconds), these values can be recalculated by
+            # setting `calc_lst=True` when calling read_miriad.
+            lst = uv["lst"]
             inttime = uv["inttime"]
-            source = uv["source"]
-            if source != _source:
-                if skip_extra_sources:
-                    if warn_extra_sources:
-                        warnings.warn(
-                            "File containts more than one source, only using data "
-                            "where source = %s" % _source
-                        )
-                        warn_extra_sources = False
-                    continue
-                raise NotImplementedError(
-                    "This appears to be a multi source file, which is not supported. "
-                    "You can bypass this error bu using skip_extra_sources=True to "
-                    "read in only the data belonging to the first source in the data "
-                    "set (and skip data from other sources)."
-                )
-            else:
-                _source = source
-
-            # check extra variables for changes compared with initial value
-            for extra_variable in list(check_variables.keys()):
-                if type(check_variables[extra_variable]) == str:
-                    if uv[extra_variable] != check_variables[extra_variable]:
-                        check_variables.pop(extra_variable)
-                else:
-                    if not np.allclose(
-                        uv[extra_variable], check_variables[extra_variable]
-                    ):
-                        check_variables.pop(extra_variable)
+            try:
+                sou_id = sou_dict[uv["source"]]
+            except KeyError:
+                sou_dict[uv["source"]] = Nphase
+                sou_id = Nphase
+                Nphase += 1
 
             try:
+                # TODO: Gotta clean this up at some point - there are going to be
+                # addition things to stick into the metadata accumulator, and accessing
+                # those values by index number will likely make it more difficult to
+                # maintain/expand the code in the future.
                 data_accumulator[uv["pol"]].append(
-                    [uvw, t, i, j, d, f, cnt, ra, dec, inttime]
+                    [
+                        uvw,  # Entry 0
+                        t,  # Entry 1
+                        i,  # Entry 2
+                        j,  # Entry 3
+                        d,  # Entry 4
+                        f,  # Entry 5
+                        cnt,  # Entry 6
+                        ra,  # Entry 7
+                        dec,  # Entry 8
+                        inttime,  # Entry 9
+                        sou_id,  # Entry 10
+                        epoch,  # Entry 11
+                        app_ra,  # Entry 12
+                        app_dec,  # Entry 13
+                        frame_pa,  # Entry 14
+                        lst,  # Entry 15
+                    ]
                 )
             except (KeyError):
                 data_accumulator[uv["pol"]] = [
-                    [uvw, t, i, j, d, f, cnt, ra, dec, inttime]
+                    [
+                        uvw,  # Entry 0
+                        t,  # Entry 1
+                        i,  # Entry 2
+                        j,  # Entry 3
+                        d,  # Entry 4
+                        f,  # Entry 5
+                        cnt,  # Entry 6
+                        ra,  # Entry 7
+                        dec,  # Entry 8
+                        inttime,  # Entry 9
+                        sou_id,  # Entry 10
+                        epoch,  # Entry 11
+                        app_ra,  # Entry 12
+                        app_dec,  # Entry 13
+                        frame_pa,  # Entry 14
+                        lst,  # Entry 15
+                    ]
                 ]
                 pol_list.append(uv["pol"])
                 # NB: flag types in miriad are usually ints
@@ -1213,7 +1286,7 @@ class Miriad(UVData):
         # instead of _miriad values which come from pyephem.
         # The differences are of order 5 seconds.
         proc = None
-        if self.telescope_location is not None:
+        if (self.telescope_location is not None) and calc_lst:
             proc = self.set_lsts_from_time_array(background=background_lsts)
         self.nsample_array = np.ones(self.data_array.shape, dtype=np.float64)
 
@@ -1221,6 +1294,13 @@ class Miriad(UVData):
         ra_pol_list = np.zeros((self.Nblts, self.Npols))
         dec_pol_list = np.zeros((self.Nblts, self.Npols))
         uvw_pol_list = np.zeros((self.Nblts, 3, self.Npols))
+        sou_id_pol_list = np.zeros((self.Nblts, self.Npols), dtype=int)
+        epoch_pol_list = np.zeros((self.Nblts, self.Npols))
+        app_ra_pol_list = np.zeros((self.Nblts, self.Npols))
+        app_dec_pol_list = np.zeros((self.Nblts, self.Npols))
+        frame_pa_pol_list = np.zeros((self.Nblts, self.Npols))
+        lst_pol_list = np.zeros((self.Nblts, self.Npols))
+
         c_ns = const.c.to("m/ns").value
         for pol, data in data_accumulator.items():
             pol_ind = self._pol_to_ind(pol)
@@ -1246,19 +1326,41 @@ class Miriad(UVData):
                 uvw_pol_list[blt_index, :, pol_ind] = uvw
                 ra_pol_list[blt_index, pol_ind] = d[7]
                 dec_pol_list[blt_index, pol_ind] = d[8]
+                sou_id_pol_list[blt_index, pol_ind] = d[10]
+                epoch_pol_list[blt_index, pol_ind] = d[11]
+                app_ra_pol_list[blt_index, pol_ind] = d[12]
+                app_dec_pol_list[blt_index, pol_ind] = d[13]
+                frame_pa_pol_list[blt_index, pol_ind] = d[14]
+                lst_pol_list[blt_index, pol_ind] = d[15]
 
         # Collapse pol axis for ra_list, dec_list, and uvw_list
         ra_list = np.zeros(self.Nblts)
         dec_list = np.zeros(self.Nblts)
+        sou_id_list = np.zeros(self.Nblts)
+        epoch_list = np.zeros(self.Nblts)
+        app_ra_list = np.zeros(self.Nblts)
+        app_dec_list = np.zeros(self.Nblts)
+        frame_pa_list = np.zeros(self.Nblts)
+        lst_list = np.zeros(self.Nblts)
+
         for blt_index in range(self.Nblts):
             test = ~np.all(self.flag_array[blt_index, :, :, :], axis=(0, 1))
             good_pol = np.where(test)[0]
+            if len(good_pol) == 0:
+                # No good pols for this blt. Fill with first one.
+                good_pol = np.array([0])
             if len(good_pol) == 1:
                 # Only one good pol, use it
                 self.uvw_array[blt_index, :] = uvw_pol_list[blt_index, :, good_pol]
                 ra_list[blt_index] = ra_pol_list[blt_index, good_pol]
                 dec_list[blt_index] = dec_pol_list[blt_index, good_pol]
-            elif len(good_pol) > 1:
+                sou_id_list[blt_index] = sou_id_pol_list[blt_index, good_pol]
+                epoch_list[blt_index] = epoch_pol_list[blt_index, good_pol]
+                app_ra_list[blt_index] = app_ra_pol_list[blt_index, good_pol]
+                app_dec_list[blt_index] = app_dec_pol_list[blt_index, good_pol]
+                frame_pa_list[blt_index] = frame_pa_pol_list[blt_index, good_pol]
+                lst_list[blt_index] = lst_pol_list[blt_index, good_pol]
+            else:
                 # Multiple good pols, check for consistency. pyuvdata does not
                 # support pol-dependent uvw, ra, or dec.
                 if np.any(np.diff(uvw_pol_list[blt_index, :, good_pol], axis=0)):
@@ -1267,19 +1369,30 @@ class Miriad(UVData):
                     self.uvw_array[blt_index, :] = uvw_pol_list[
                         blt_index, :, good_pol[0]
                     ]
-                if np.any(np.diff(ra_pol_list[blt_index, good_pol])):
-                    raise ValueError("ra values are different by polarization.")
-                else:
-                    ra_list[blt_index] = ra_pol_list[blt_index, good_pol[0]]
-                if np.any(np.diff(dec_pol_list[blt_index, good_pol])):
-                    raise ValueError("dec values are different by polarization.")
-                else:
-                    dec_list[blt_index] = dec_pol_list[blt_index, good_pol[0]]
-            else:
-                # No good pols for this blt. Fill with first one.
-                self.uvw_array[blt_index, :] = uvw_pol_list[blt_index, :, 0]
-                ra_list[blt_index] = ra_pol_list[blt_index, 0]
-                dec_list[blt_index] = dec_pol_list[blt_index, 0]
+
+                check_list = [ra_pol_list, dec_pol_list, sou_id_pol_list, lst_pol_list]
+                assign_list = [ra_list, dec_list, sou_id_list, lst_list]
+                check_names = ["ra", "dec", "source id", "lst"]
+                if record_epoch:
+                    check_list.append(epoch_pol_list)
+                    assign_list.append(epoch_list)
+                    check_names.append("epoch")
+                if record_app:
+                    check_list.extend([app_ra_pol_list, app_dec_pol_list])
+                    assign_list.extend([app_ra_list, app_dec_list])
+                    check_names.extend(["obsra", "obsdec"])
+                if record_pa:
+                    check_list.append(frame_pa_pol_list)
+                    assign_list.append(frame_pa_list)
+                    check_names.append("obspa")
+
+                for item, name in zip(check_list, check_names):
+                    if np.any(np.diff(item[blt_index, good_pol])):
+                        raise ValueError(
+                            "%s values are different by polarization." % name
+                        )
+                for item, target in zip(check_list, assign_list):
+                    target[blt_index] = item[blt_index, good_pol[0]]
 
         # get unflagged blts
         # If we have a 1-baseline, single integration data set, set single_ra and
@@ -1300,6 +1413,8 @@ class Miriad(UVData):
         if phase_type is not None:
             if phase_type == "phased":
                 self._set_phased()
+                if Nphase > 1:
+                    self._set_multi_phase_center()
             elif phase_type == "drift":
                 self._set_drift()
             else:
@@ -1309,12 +1424,15 @@ class Miriad(UVData):
                     "reflect the phasing status of the data"
                 )
         else:
-            # check if ra is constant throughout file; if it is,
-            # file is tracking if not, file is drift scanning
-            # check if there's only one unflagged time
+            # check if ra is constant throughout file OR if there are multiple sources
+            # listed; if the obs meets either of those conditions, then assume that the
+            # file is tracking, and if not, assume that the  file is drift scanning,
+            # checking if there's only one unflagged time
             if not single_time:
-                if single_ra:
+                if single_ra or (Nphase > 1):
                     self._set_phased()
+                    if Nphase > 1:
+                        self._set_multi_phase_center()
                 else:
                     self._set_drift()
             else:
@@ -1329,7 +1447,50 @@ class Miriad(UVData):
         if proc is not None:
             proc.join()
 
-        if self.phase_type == "phased":
+        if (self.telescope_location is None) or not calc_lst:
+            # The float below is the number of sidereal days per solar day, and the
+            # formula below adjusts for the fact that in MIRIAD, the lst is for the
+            # start of the integration, as opposed to pyuvdata, which evaluates these
+            # values at the midpoint of the integration.
+            self.lst_array = lst_list + (
+                np.pi * 1.002737909350795 * self.integration_time / (24.0 * 3600.0)
+            )
+
+        if record_app and (self.phase_type == "phased"):
+            self.phase_center_app_ra = app_ra_list
+            self.phase_center_app_dec = app_dec_list
+            if record_pa:
+                self.phase_center_frame_pa = frame_pa_list
+
+        if self.multi_phase_center:
+            # This presupposes that the data are already phased
+            self.phase_center_ra = 0.0  # This isn't used for mutli-phase-ctr data
+            self.phase_center_dec = 0.0  # This isn't used for mutli-phase-ctr data
+            # The concept of an epoch is kinda weird in mutli-phase-ctr datasets, b/c
+            # different sources are allowed to have different epochs. Just set it to
+            # the median value for now. Eventually, I think this should be used as the
+            # "if your format forces me to use FK5, here's the epoch" value. Note that
+            # MIRIAD, AIPS, and (I think) CASA tasks assume the coordinates are given
+            # in FK5 format (well, unless epoch <= 1984, in which case MIRIAD assumes
+            # in semi-Orwellian fashion that you _really_ wanted FK4/Bessel-Newcomb).
+            self.phase_center_epoch = 2000.0
+            self.phase_center_frame = "icrs"
+            self.phase_center_id_array = sou_id_list.astype(int)
+            # Here is where we should package up sources
+            for name in sou_dict.keys():
+                select_mask = sou_id_list == sou_dict[name]
+                epoch_val = np.median(epoch_list[select_mask])
+                self._add_phase_center(
+                    name,
+                    "sidereal",
+                    cat_lon=np.median(ra_list[select_mask]),
+                    cat_lat=np.median(dec_list[select_mask]),
+                    cat_frame="fk4" if (epoch_val < 1984.0) else "fk5",
+                    cat_epoch=epoch_val,
+                    cat_id=sou_dict[name],
+                    info_source="file",
+                )
+        elif self.phase_type == "phased":
             # check that the RA values do not vary
             if not single_ra:
                 raise ValueError(
@@ -1338,8 +1499,11 @@ class Miriad(UVData):
             self.phase_center_ra = float(ra_list[0])
             self.phase_center_dec = float(dec_list[0])
             self.phase_center_epoch = uv["epoch"]
-            if "phsframe" in uv.vartable.keys():
+            if "phsframe" not in uv.vartable.keys():
+                self.phase_center_frame = "fk4" if (uv["epoch"] < 1984.0) else "fk5"
+            else:
                 self.phase_center_frame = uv["phsframe"].replace("\x00", "")
+
         else:
             # check that the RA values are not constant (if more than one time present)
             if single_ra and not single_time:
@@ -1367,13 +1531,19 @@ class Miriad(UVData):
                 warnings.warn(
                     "drift RA, Dec is off from lst, latitude by more than {}, "
                     "so it appears that it is not a zenith drift scan. "
-                    'Setting phase_type to "unknown"'.format(acceptable_offset)
+                    'Setting phase_type to "drift" for now'.format(acceptable_offset)
                 )
-                self._set_unknown_phase_type()
+                self._set_drift()
 
         # close out now that we're done
         uv.close()
 
+        if not (record_app and record_pa) and (self.phase_type == "phased"):
+            # At this point, if we are missing information about the sky position of the
+            # source, we want to fill it in now. If we have the apparent coords, but are
+            # only missing the frame position angles (common given that obspa is not a
+            # standard keyword), then we can _just_ fill those in.
+            self._set_app_coords_helper(pa_only=record_app)
         try:
             self.set_telescope_params()
         except ValueError as ve:
@@ -1389,12 +1559,26 @@ class Miriad(UVData):
                 minor_order = None
             self.reorder_blts(order=order, minor_order=minor_order)
 
+        # If the data set was recorded using the old phasing method, fix that now.
+        if fix_old_proj and (self.phase_type == "phased"):
+            if not self.multi_phase_center:
+                self.fix_phase(use_ant_pos=fix_use_ant_pos)
+            else:
+                warnings.warn(
+                    "Cannot fix the phases of multi phase center datasets, as they "
+                    "were not supported when the old phasing method was used, and "
+                    "thus, there is no need to correct the data."
+                )
+
         # check if object has all required uv_properties set
         if run_check:
             self.check(
                 check_extra=check_extra,
                 run_check_acceptability=run_check_acceptability,
                 strict_uvw_antpos_check=strict_uvw_antpos_check,
+                allow_flip_conj=True,
+                check_autos=check_autos,
+                fix_autos=fix_autos,
             )
 
     def write_miriad(
@@ -1406,6 +1590,9 @@ class Miriad(UVData):
         run_check_acceptability=True,
         strict_uvw_antpos_check=False,
         no_antnums=False,
+        calc_lst=False,
+        check_autos=True,
+        fix_autos=False,
     ):
         """
         Write the data to a miriad file.
@@ -1433,13 +1620,25 @@ class Miriad(UVData):
         no_antnums : bool
             Option to not write the antnums variable to the file.
             Should only be used for testing purposes.
+        calc_lst : bool
+            Recalculate the LST values upon writing the file. This is done to perform
+            higher-precision accounting for the difference in MIRAD timestamps vs
+            pyuvdata (the former marks the beginning of an integration, the latter
+            marks the midpoint). Default is False, which instead uses a simple formula
+            for correcting the LSTs, expected to be accurate to approximately 0.1 µsec
+            precision.
+        check_autos : bool
+            Check whether any auto-correlations have non-zero imaginary values in
+            data_array (which should not mathematically exist). Default is True.
+        fix_autos : bool
+            If auto-correlations with imaginary values are found, fix those values so
+            that they are real-only in data_array. Default is False.
 
         Raises
         ------
         ValueError
             If the frequencies are not evenly spaced or are separated by more
             than their channel width.
-            The `phase_type` of the object is "unknown".
         TypeError
             If any entry in extra_keywords is not a single string or number.
 
@@ -1449,11 +1648,17 @@ class Miriad(UVData):
         # change time_array and lst_array to mark beginning of integration,
         # per Miriad format
         miriad_time_array = self.time_array - self.integration_time / (24 * 3600.0) / 2
-        if self.telescope_location is not None:
+        if (self.telescope_location is not None) and calc_lst:
             latitude, longitude, altitude = self.telescope_location_lat_lon_alt_degrees
             miriad_lsts = uvutils.get_lst_for_time(
                 miriad_time_array, latitude, longitude, altitude
             )
+        else:
+            # The long float below is the number of sidereal days per day. The below
+            # equation should be accurate to _much_ better than 1 µsec.
+            miriad_lsts = self.lst_array - (
+                np.pi * 1.002737909350795 * self.integration_time / (24.0 * 3600.0)
+            )
 
         # Miriad requires j>i which we call ant1<ant2
         self.conjugate_bls(convention="ant1<ant2")
@@ -1464,6 +1669,8 @@ class Miriad(UVData):
                 run_check_acceptability=run_check_acceptability,
                 check_freq_spacing=True,
                 strict_uvw_antpos_check=strict_uvw_antpos_check,
+                check_autos=check_autos,
+                fix_autos=fix_autos,
             )
 
         if os.path.exists(filepath):
@@ -1544,8 +1751,9 @@ class Miriad(UVData):
 
         # NB: restfreq should go in here at some point
         #####################################################
-        uv.add_var("source", "a")
-        uv["source"] = self.object_name
+        if not self.multi_phase_center:
+            uv.add_var("source", "a")
+            uv["source"] = self.object_name
         uv.add_var("telescop", "a")
         uv["telescop"] = self.telescope_name
         uv.add_var("latitud", "d")
@@ -1601,8 +1809,9 @@ class Miriad(UVData):
             )
 
         if self.phase_type == "phased":
-            uv.add_var("epoch", "r")
-            uv["epoch"] = self.phase_center_epoch
+            if self.phase_center_epoch is not None:
+                uv.add_var("epoch", "r")
+                uv["epoch"] = self.phase_center_epoch
             if self.phase_center_frame is not None:
                 uv.add_var("phsframe", "a")
                 uv["phsframe"] = self.phase_center_frame
@@ -1651,20 +1860,6 @@ class Miriad(UVData):
         # vectors.  This is not supported at present!
         # NB: complex numbers *should* be supportable, but are not currently
         # supported due to unexplained errors in _miriad and/or its underlying libraries
-        numpy_types = {
-            np.int8: int,
-            np.int16: int,
-            np.int32: int,
-            np.int64: int,
-            np.uint8: int,
-            np.uint16: int,
-            np.uint32: int,
-            np.uint64: int,
-            np.float16: float,
-            np.float32: float,
-            np.float64: float,
-            np.float128: float,
-        }
         types = {
             str: "a",
             int: "i",
@@ -1672,14 +1867,20 @@ class Miriad(UVData):
             bool: "a",  # booleans are stored as strings and changed back on read
         }
         for key, value in self.extra_keywords.items():
-            if isinstance(value, tuple(numpy_types.keys())):
-                if numpy_types[type(value)] == int:
+            raise_type_error = False
+            if isinstance(value, np.number):
+                if issubclass(value.dtype.type, np.integer):
                     value = int(value)
-                elif numpy_types[type(value)] == float:
+                elif issubclass(value.dtype.type, np.floating):
                     value = float(value)
+                elif issubclass(value.dtype.type, np.complexfloating):
+                    raise_type_error = True
             elif type(value) == bool:
                 value = str(value)
             elif type(value) not in types.keys():
+                raise_type_error = True
+
+            if raise_type_error:
                 raise TypeError(
                     "Extra keyword {keyword} is of {keytype}. "
                     "Only strings and real numbers are "
@@ -1716,10 +1917,27 @@ class Miriad(UVData):
         uv.add_var("pol", "i")
         uv.add_var("lst", "d")
         uv.add_var("cnt", "d")
+        if self.multi_phase_center:
+            uv.add_var("source", "a")
+            name_dict = {
+                self.phase_center_catalog[key]["cat_id"]: key
+                for key in self.phase_center_catalog.keys()
+            }
         uv.add_var("ra", "d")
         uv.add_var("dec", "d")
         uv.add_var("inttime", "d")
 
+        app_record = False
+        if (
+            self.phase_center_app_ra is not None
+            and self.phase_center_app_dec is not None
+            and self.phase_center_frame_pa is not None
+        ):
+            app_record = True
+            uv.add_var("obsra", "d")
+            uv.add_var("obsdec", "d")
+            uv.add_var("obspa", "d")  # Non-standard MIRIAD keyword
+
         # write data
         c_ns = const.c.to("m/ns").value
         for viscnt, blt in enumerate(self.data_array):
@@ -1731,19 +1949,22 @@ class Miriad(UVData):
             uv["lst"] = miriad_lsts[viscnt].astype(np.double)
             uv["inttime"] = self.integration_time[viscnt].astype(np.double)
             if self.phase_type == "phased":
-                uv["ra"] = self.phase_center_ra
-                uv["dec"] = self.phase_center_dec
+                if self.multi_phase_center:
+                    cat_name = name_dict[self.phase_center_id_array[viscnt]]
+                    uv["source"] = cat_name
+                    uv["ra"] = self.phase_center_catalog[cat_name]["cat_lon"]
+                    uv["dec"] = self.phase_center_catalog[cat_name]["cat_lat"]
+                else:
+                    uv["ra"] = self.phase_center_ra
+                    uv["dec"] = self.phase_center_dec
+                if app_record:
+                    uv["obsra"] = self.phase_center_app_ra[viscnt]
+                    uv["obsdec"] = self.phase_center_app_dec[viscnt]
+                    uv["obspa"] = self.phase_center_frame_pa[viscnt]
             elif self.phase_type == "drift":
                 uv["ra"] = miriad_lsts[viscnt].astype(np.double)
                 uv["dec"] = self.telescope_location_lat_lon_alt[0].astype(np.double)
-            else:
-                raise ValueError(
-                    "The phasing type of the data is unknown. "
-                    'Set the phase_type to "drift" or "phased" to '
-                    "reflect the phasing status of the data"
-                )
 
-            # NOTE only writing spw 0, not supporting multiple spws for write
             for polcnt, pol in enumerate(self.polarization_array):
                 uv["pol"] = pol.astype(np.int64)
                 if self.future_array_shapes:
